from typing import List

import numpy as np
from torch.utils.data import Dataset

from .strategy import Strategy


class RandomSampling(Strategy):
    """Random Sampling Class.

    Args:
        pool_images (List[str]): List of pool image paths.
        pool_labels (List[str]): List of pool label paths.
        val_images (List[str]): List of validation image paths.
        val_labels (List[str]): List of validation label paths.
        test_images (List[str]): List of test image paths.
        test_labels (List[str]): List of test label paths.
        idxs_lb (np.ndarray): Array of bool type to record labeled data.
        model_params (dict): Model parameters.
                            e.g. model_params = {
                                    "MODEL_NAME": MODEL_NAME,
                                    "ENCODER": ENCODER,
                                    "ENCODER_WEIGHTS": ENCODER_WEIGHTS,
                                    "NUM_CLASSES": NUM_CLASSES,
                                }
        dataset (Dataset): Dataset class.
        dataset_params (dict): Dataset parameters.
                            e.g.     dataset_params = {
                                            "training_augmentation": get_training_augmentation(),
                                            "validation_augmentation": get_validation_augmentation(),
                                            "preprocessing": get_preprocessing(preprocessing_fn),
                                            "classes": CamvidDataset.CLASSES,
                                        }
    """

    def __init__(
        self,
        pool_images: List[str],
        pool_labels: List[str],
        val_images: List[str],
        val_labels: List[str],
        test_images: List[str],
        test_labels: List[str],
        idxs_lb: np.ndarray,
        model_params: dict,
        dataset: Dataset,
        dataset_params: dict,
    ):
        super(RandomSampling, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model_params,
            dataset,
            dataset_params,
        )

    def query(self, n: int) -> List[int]:
        """Query data.

        Args:
            n (int): Number of data to query.

        Returns:
            List[int]: Indices of queried data.
        """
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        np.random.shuffle(idxs_unlabeled)
        idxs_queried = idxs_unlabeled[:n]  # idxs_queried: index in pool_images
        return idxs_queried



import json
from typing import Any, List

import numpy as np


def is_list_of_strings(lst: List[str]) -> bool:
    """Check if a list only contains strings

    Args:
        lst (List[str]): a list

    Returns:
        bool: if all element is string, return True, otherwise return False
    """
    if lst and isinstance(lst, list):
        return all(isinstance(elem, str) for elem in lst)
    else:
        return False


def is_list_of_int(lst: List[int]) -> bool:
    """Check if a list only contains int

    Args:
        lst (List[int]): a list

    Returns:
        bool: if all element is int, return True, otherwise return False
    """
    if lst and isinstance(lst, list):
        return all(isinstance(elem, int) for elem in lst)
    else:
        return False


def is_array_of_bools(array: np.ndarray) -> bool:
    """Check if a array only contains bool

    Args:
        array (np.ndarray): numpy array

    Returns:
        bool: if all element types are bool, return True, otherwise return False
    """
    if isinstance(array, np.ndarray) and len(list(array)) != []:
        return all(type(elem) != bool for elem in array)
    else:
        return False


def save_json(data: Any, save_path: str) -> None:
    """Save data to json format.
    Args:
        data (Any): The data to store.
        save_path (str): Path to save.
    """
    with open(save_path, "w", encoding="utf-8") as f:
        json.dump(data, f, indent=2, ensure_ascii=False)


def load_json(path: str) -> List[Any]:
    """Read json data
    Args:
        path (str): Path of file to read.
    Returns:
        List[Any]: FTSE entity data.
    """
    with open(path, "r", encoding="utf-8") as f:
        data = json.load(f)
    return data



from typing import List, Optional, Tuple

import cv2
import numpy as np
from albumentations import BaseCompose
from torch import Tensor
from torch.utils.data import Dataset

from segal.utils import is_list_of_int, is_list_of_strings


class BaseDataset(Dataset):
    """Base dataset. Read images, apply augmentation and preprocessing transformations.

    Args:
        image_paths (List[str]): path to images
        mask_paths (List[str]): path to segmentation masks
        class_values ([List[int]]): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. noralization, shape manipulation, etc.)
    """

    def __init__(
        self,
        image_paths: List[str],
        mask_paths: List[str],
        class_values: List[int],
        augmentation: Optional[BaseCompose] = None,
        preprocessing: Optional[BaseCompose] = None,
    ):
        if not all([is_list_of_strings(image_paths), is_list_of_strings(mask_paths)]):
            raise TypeError("Images paths must be a list of string!")
        if not is_list_of_int(class_values):
            raise TypeError("classes must be a numpy array!")
        self.image_paths = image_paths
        self.mask_paths = mask_paths
        self.class_values = class_values
        self.augmentation = augmentation
        self.preprocessing = preprocessing

    def __getitem__(self, i) -> Tuple[Tensor, Tensor]:
        """Get data by index.

        Args:
            i (int): Index.

        Returns:
            Tensor, Tensor: image and mask tensor.
        """

        # read data
        image = cv2.imread(self.image_paths[i])  # Read as BGR, 0~255
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # Convert to RGB, 0~1
        mask = cv2.imread(self.mask_paths[i], 0)

        # extract certain classes from mask (e.g. cars)
        masks = [(mask == v) for v in self.class_values]
        mask = np.stack(masks, axis=-1).astype("float")

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        # apply preprocessing
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def __len__(self) -> int:
        """return size of dataset.

        Returns:
            int: size of dataset.
        """
        return len(self.image_paths)



from typing import List, Optional, Tuple

import cv2
import numpy as np
from albumentations import BaseCompose
from torch import Tensor

from segal.datasets.base_dataset import BaseDataset


class CityscapesDataset(BaseDataset):
    """Cityscapes Dataset.

    Args:
        image_paths (List[str]): path to images
        mask_paths (List[str]): path to segmentation masks
        class_values ([List[int]]): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. noralization, shape manipulation, etc.)
    """

    mapping = {
        0: 0,
        1: 0,
        2: 0,
        3: 0,
        4: 0,
        5: 0,
        6: 0,
        7: 1,
        8: 2,
        9: 0,
        10: 0,
        11: 3,
        12: 4,
        13: 5,
        14: 0,
        15: 0,
        16: 0,
        17: 6,
        18: 0,
        19: 7,
        20: 8,
        21: 9,
        22: 10,
        23: 11,
        24: 12,
        25: 13,
        26: 14,
        27: 15,
        28: 16,
        29: 0,
        30: 0,
        31: 17,
        32: 18,
        33: 19,
        -1: 0,
    }

    # The values above are remapped to the following
    CLASSES = [
        "unlabeled",
        "road",
        "sidewalk",
        "building",
        "wall",
        "fence",
        "pole",
        "traffic_light",
        "traffic_sign",
        "vegetation",
        "terrain",
        "sky",
        "person",
        "rider",
        "car",
        "truck",
        "bus",
        "train",
        "motorcycle",
        "bicycle",
    ]

    def __init__(
        self,
        image_paths: List[str],
        mask_paths: List[str],
        class_values: List[int],
        augmentation: Optional[BaseCompose] = None,
        preprocessing: Optional[BaseCompose] = None,
    ):
        super(CityscapesDataset, self).__init__(
            image_paths,
            mask_paths,
            class_values,
            augmentation,
            preprocessing,
        )

    def __getitem__(self, i) -> Tuple[Tensor, Tensor]:
        """Get data by index.

        Args:
            i (int): Index.

        Returns:
            Tensor, Tensor: image and mask tensor.
        """

        # read data
        image = cv2.imread(self.image_paths[i])  # Read as BGR, 0~255
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)  # Convert to RGB, 0~1
        mask = cv2.imread(self.mask_paths[i], 0)

        # Remap label
        mask = self.encode_labels(mask)
        # extract certain classes from mask (e.g. cars)
        masks = [(mask == v) for v in self.class_values]
        mask = np.stack(masks, axis=-1).astype("float")

        # apply augmentations
        if self.augmentation:
            sample = self.augmentation(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        # apply preprocessing
        if self.preprocessing:
            sample = self.preprocessing(image=image, mask=mask)
            image, mask = sample["image"], sample["mask"]

        return image, mask

    def encode_labels(self, mask: np.ndarray) -> np.ndarray:
        """Convert 33 classes to 20 classes

        Args:
            mask (np.ndarray): mask with 33 classes id

        Returns:
            np.ndarray: mask with 20 classes id
        """
        label_mask = np.zeros_like(mask)
        for k in self.mapping:
            label_mask[mask == k] = self.mapping[k]
        return label_mask



from typing import List, Optional

from albumentations import BaseCompose

from segal.datasets.base_dataset import BaseDataset


class CamvidDataset(BaseDataset):
    """CamVid Dataset. Read images, apply augmentation and preprocessing transformations.

    Args:
        image_paths (List[str]): path to images
        mask_paths (List[str]): path to segmentation masks
        class_values ([List[int]]): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. noralization, shape manipulation, etc.)
    """

    CLASSES = [
        "sky",
        "building",
        "pole",
        "road",
        "pavement",
        "tree",
        "signsymbol",
        "fence",
        "car",
        "pedestrian",
        "bicyclist",
        "unlabelled",
    ]

    def __init__(
        self,
        image_paths: List[str],
        mask_paths: List[str],
        class_values: List[int],
        augmentation: Optional[BaseCompose] = None,
        preprocessing: Optional[BaseCompose] = None,
    ):
        super(CamvidDataset, self).__init__(
            image_paths,
            mask_paths,
            class_values,
            augmentation,
            preprocessing,
        )



from pathlib import Path
from typing import List, Optional

import numpy as np
import segmentation_models_pytorch as smp
import torch
import torch.nn.functional as F
from segmentation_models_pytorch import utils
from torch.utils.data import DataLoader, Dataset

from segal.utils import is_array_of_bools, is_list_of_strings


class Strategy:
    """Base Strategy class.

    Attributes:
        pool_images (List[str]): List of pool image paths.
        pool_labels (List[str]): List of pool label paths.
        val_images (List[str]): List of validation image paths.
        val_labels (List[str]): List of validation label paths.
        test_images (List[str]): List of test image paths.
        test_labels (List[str]): List of test label paths.
        idxs_lb (np.ndarray): Array of bool type to record labeled data.
        model_params (dict): Model parameters.
        model_params (dict): Model parameters.
        dataset (Dataset): Dataset class.
        dataset_params (dict): Dataset parameters.
        val_dataset (Union[Dataset, None]): Validation dataset.
        test_dataset (Union[Dataset, None]): Validation dataset.
        n_pool (int): Number of pool data.
        device (str): "cuda" or "cpu".
        best_model (torch.nn.Module): Segmentation model.
        train_logs (List[dict]): List of training logs.
        val_logs (List[dict]): List of validation logs.
        test_logs (List[dict]): List of test logs.
    """

    def __init__(
        self,
        pool_images: List[str],
        pool_labels: List[str],
        val_images: List[str],
        val_labels: List[str],
        test_images: List[str],
        test_labels: List[str],
        idxs_lb: np.ndarray,
        model_params: dict,
        dataset: Dataset,
        dataset_params: dict,
    ):
        """Initialize Strategy class.

        Args:
            pool_images (List[str]): List of pool image paths.
            pool_labels (List[str]): List of pool label paths.
            val_images (List[str]): List of validation image paths.
            val_labels (List[str]): List of validation label paths.
            test_images (List[str]): List of test image paths.
            test_labels (List[str]): List of test label paths.
            idxs_lb (np.ndarray): Array of bool type to record labeled data.
            model_params (dict): Model parameters.
                                e.g. model_params = {
                                        "MODEL_NAME": MODEL_NAME,
                                        "ENCODER": ENCODER,
                                        "ENCODER_WEIGHTS": ENCODER_WEIGHTS,
                                        "NUM_CLASSES": NUM_CLASSES,
                                    }
            dataset (Dataset): Dataset class.
            dataset_params (dict): Dataset parameters.
                                e.g.     dataset_params = {
                                                "training_augmentation": get_training_augmentation(),
                                                "validation_augmentation": get_validation_augmentation(),
                                                "preprocessing": get_preprocessing(preprocessing_fn),
                                                "classes": CamvidDataset.CLASSES,
                                            }
        """
        if not all(
            [
                is_list_of_strings(pool_images),
                is_list_of_strings(pool_labels),
                is_list_of_strings(val_images),
                is_list_of_strings(val_labels),
                is_list_of_strings(test_images),
                is_list_of_strings(test_images),
                is_list_of_strings(test_labels),
            ]
        ):
            raise TypeError("Images paths must be a list of string!")
        if not is_array_of_bools(idxs_lb):
            raise TypeError("idxs_lb must be a numpy array of bool!")
        if not isinstance(dataset_params, dict):
            raise TypeError("dataset_params must be a dict!")
        if not isinstance(model_params, dict):
            raise TypeError("model_params must be a numpy array!")

        self.pool_images: List[str] = pool_images
        self.pool_labels: List[str] = pool_labels
        self.val_images: List[str] = val_images
        self.val_labels: List[str] = val_labels
        self.test_images: List[str] = test_images
        self.test_labels: List[str] = test_labels
        self.idxs_lb: np.ndarray = idxs_lb
        self.dataset: Dataset = dataset
        self.dataset_params: dict = dataset_params
        self.model_params: dict = model_params
        self.val_dataset = None
        self.test_dataset = None
        self.n_pool = len(pool_images)
        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.best_model: torch.nn.Module = torch.nn.Module()
        self.train_logs: List[dict] = []
        self.val_logs: List[dict] = []
        self.test_logs: List[dict] = []

    def query(self, n: int) -> List[int]:
        """Query data

        Args:
            n (int): num of query data

        Returns:
            List[int]: index of query data
        """
        pass

    def update(self, idxs_lb: np.ndarray) -> None:
        """Update labeled data index

        Args:
            idxs_lb (np.ndarray): array of bool type to record labeled data.
        """
        self.idxs_lb = idxs_lb

    def train(
        self,
        n_epoch: int = 10,
        activation: str = "softmax2d",
        save_path: str = "output",
    ) -> dict:
        """Train model and return performance on test data

        Args:
            n_epoch (int, optional): num of epochs. Defaults to 10.
            activation (str, optional): activation function. Defaults to "softmax2d".
            save_path (str, optional): save path of result. Defaults to "output".

        Returns:
            dict: a dict with log of performance
        """
        model = smp.__dict__[self.model_params["MODEL_NAME"]](
            encoder_name=self.model_params["ENCODER"],
            encoder_weights=self.model_params["ENCODER_WEIGHTS"],
            classes=self.model_params["NUM_CLASSES"],
        )
        model.to(self.device)

        base_path = Path(save_path)
        if not base_path.exists():
            base_path.mkdir(exist_ok=True, parents=True)

        idxs_train = np.arange(self.n_pool)[self.idxs_lb]

        train_images = [self.pool_images[idx] for idx in idxs_train]
        train_labels = [self.pool_labels[idx] for idx in idxs_train]

        train_dataset: Dataset = self.dataset(
            train_images,
            train_labels,
            augmentation=self.dataset_params["training_augmentation"],
            preprocessing=self.dataset_params["preprocessing"],
            class_values=self.dataset_params["class_values"],
        )  # type: ignore[operator]

        if self.val_dataset is None:
            valid_dataset = self.dataset(
                self.val_images,
                self.val_labels,
                augmentation=self.dataset_params["validation_augmentation"],
                preprocessing=self.dataset_params["preprocessing"],
                class_values=self.dataset_params["class_values"],
            )  # type: ignore[operator]
            self.valid_dataset = valid_dataset
        else:
            valid_dataset = self.valid_dataset

        if self.device == "cuda":
            train_loader = DataLoader(
                train_dataset, batch_size=4, shuffle=True, num_workers=4
            )
            valid_loader = DataLoader(
                valid_dataset, batch_size=1, shuffle=False, num_workers=2
            )
        else:
            train_loader = DataLoader(train_dataset, batch_size=4, shuffle=True)
            valid_loader = DataLoader(valid_dataset, batch_size=1, shuffle=False)

        loss = utils.losses.DiceLoss(activation=activation)
        metrics = [
            utils.metrics.IoU(threshold=0.5, activation=activation),
        ]

        optimizer = torch.optim.Adam(
            [
                dict(params=model.parameters(), lr=0.0001),
            ]
        )

        train_epoch = utils.train.TrainEpoch(
            model,
            loss=loss,
            metrics=metrics,
            optimizer=optimizer,
            device=self.device,
            verbose=True,
        )

        valid_epoch = utils.train.ValidEpoch(
            model,
            loss=loss,
            metrics=metrics,
            device=self.device,
            verbose=True,
        )

        max_score = 0
        for i in range(0, n_epoch):

            print("\nEpoch: {}".format(i))
            train_logs = train_epoch.run(train_loader)
            valid_logs = valid_epoch.run(valid_loader)
            self.train_logs.append(train_logs)
            self.val_logs.append(valid_logs)

            # Save best model
            if max_score < valid_logs["iou_score"]:
                max_score = valid_logs["iou_score"]
                self.best_model = model
                model_path = base_path / "best_model.pth"
                torch.save(model, model_path)
                print("Model saved!")

        return valid_logs

    def evaluate(
        self, activation: str = "softmax2d", check_point: Optional[str] = None
    ):
        """Evaluate on dataset

        Args:
            activation (str, optional): activation function. Defaults to "softmax2d".
            check_point (Optional[str], optional): saved model. Defaults to None.
        """
        loss = utils.losses.DiceLoss(activation=activation)
        metrics = [
            utils.metrics.IoU(threshold=0.5, activation=activation),
        ]
        if check_point:
            model = torch.load(check_point)
        else:
            model = self.best_model

        # Evaluate model on test data
        if self.test_dataset is None:
            test_dataset = self.dataset(
                self.test_images,
                self.test_labels,
                augmentation=self.dataset_params["validation_augmentation"],
                preprocessing=self.dataset_params["preprocessing"],
                class_values=self.dataset_params["class_values"],
            )  # type: ignore[operator]
            self.test_dataset = test_dataset
        else:
            test_dataset = self.test_dataset

        print("Evaluate on test data")
        test_dataloader = DataLoader(test_dataset, batch_size=10, shuffle=False)
        test_epoch = utils.train.ValidEpoch(
            model=model,
            loss=loss,
            metrics=metrics,
            device=self.device,
        )

        round_test_log = test_epoch.run(test_dataloader)
        self.test_logs.append(round_test_log)

        return round_test_log

    def predict_prob(self, idxs_unlabeled: List[int]) -> np.ndarray:
        """Predict on unlabeled data.

        Args:
            idxs_unlabeled (List[int]): List of unlabeled data indices.

        Returns:
            np.ndarray: Probability of classes.
        """
        model = self.best_model
        # model:torch.nn.Module = self.best_model
        unlabeled_images = [self.pool_images[idx] for idx in idxs_unlabeled]
        unlabeled_labels = [self.pool_labels[idx] for idx in idxs_unlabeled]

        unlabeled_dataset = self.dataset(
            unlabeled_images,
            unlabeled_labels,
            augmentation=self.dataset_params["validation_augmentation"],
            preprocessing=self.dataset_params["preprocessing"],
            class_values=self.dataset_params["class_values"],
        )  # type: ignore[operator]

        if self.device == "cuda":
            unlabeled_loader = DataLoader(
                unlabeled_dataset, batch_size=10, shuffle=False, num_workers=2
            )
        else:
            unlabeled_loader = DataLoader(
                unlabeled_dataset, batch_size=10, shuffle=False
            )

        probs = []
        for batch_images, _ in unlabeled_loader:
            if self.device == "cuda":
                batch_images = batch_images.to(self.device)
            out = model.predict(batch_images)  # type: ignore[operator]
            batch_probs = F.softmax(out, dim=1)
            batch_probs = batch_probs.detach().cpu().numpy()
            probs.append(batch_probs)

        concat_probs = np.concatenate(probs, axis=0)

        return concat_probs



from typing import List, Optional

from albumentations import BaseCompose

from segal.datasets.base_dataset import BaseDataset


class VOCDataset(BaseDataset):
    """VOCDataset Dataset.

    Args:
        image_paths (List[str]): path to images
        mask_paths (List[str]): path to segmentation masks
        class_values ([List[int]]): values of classes to extract from segmentation mask
        augmentation (albumentations.Compose): data transformation pipeline
            (e.g. flip, scale, etc.)
        preprocessing (albumentations.Compose): data preprocessing
            (e.g. noralization, shape manipulation, etc.)
    """

    CLASSES = [
        "background",
        "aeroplane",
        "bicycle",
        "bird",
        "boad",
        "bottle",
        "bus",
        "cat",
        "car",
        "chair",
        "cow",
        "dining table",
        "dog",
        "horse",
        "motor bike",
        "person",
        "potted plant",
        "sheep",
        "sofa",
        "train",
        "tv/monitor",
    ]

    def __init__(
        self,
        image_paths: List[str],
        mask_paths: List[str],
        class_values: List[int],
        augmentation: Optional[BaseCompose] = None,
        preprocessing: Optional[BaseCompose] = None,
    ):
        super(VOCDataset, self).__init__(
            image_paths,
            mask_paths,
            class_values,
            augmentation,
            preprocessing,
        )



from typing import List

import numpy as np
import torch
from torch.utils.data import Dataset

from .strategy import Strategy


class VoteSampling(Strategy):
    """Vote Sampling Class.

    Args:
        pool_images (List[str]): List of pool image paths.
        pool_labels (List[str]): List of pool label paths.
        val_images (List[str]): List of validation image paths.
        val_labels (List[str]): List of validation label paths.
        test_images (List[str]): List of test image paths.
        test_labels (List[str]): List of test label paths.
        idxs_lb (np.ndarray): Array of bool type to record labeled data.
        model_params (dict): Model parameters.
                            e.g. model_params = {
                                    "MODEL_NAME": MODEL_NAME,
                                    "ENCODER": ENCODER,
                                    "ENCODER_WEIGHTS": ENCODER_WEIGHTS,
                                    "NUM_CLASSES": NUM_CLASSES,
                                }
        dataset (Dataset): Dataset class.
        dataset_params (dict): Dataset parameters.
                            e.g.     dataset_params = {
                                            "training_augmentation": get_training_augmentation(),
                                            "validation_augmentation": get_validation_augmentation(),
                                            "preprocessing": get_preprocessing(preprocessing_fn),
                                            "classes": CamvidDataset.CLASSES,
                                        }
    """

    def __init__(
        self,
        pool_images: List[str],
        pool_labels: List[str],
        val_images: List[str],
        val_labels: List[str],
        test_images: List[str],
        test_labels: List[str],
        idxs_lb: np.ndarray,
        model_params: dict,
        dataset: Dataset,
        dataset_params: dict,
    ):
        super(VoteSampling, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model_params,
            dataset,
            dataset_params,
        )

    def query(self, n: int) -> List[int]:
        """Query data.

        Args:
            n (int): Number of data to query.

        Returns:
            List[int]: Indices of queried data.
        """
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        probs = self.predict_prob(idxs_unlabeled)
        scores = self.cal_scores(probs)
        topk_idxs = self.get_topk_idxs(scores, n)  # index in scores
        idxs_queried = idxs_unlabeled[topk_idxs]  # idxs_queried: index in pool_images
        return idxs_queried

    @staticmethod
    def get_topk_idxs(scores: np.ndarray, k: int) -> np.ndarray:
        """Get top k indices.

        Args:
            scores (np.ndarray): scores of batch data
            k (int): num of data to query

        Returns:
            np.ndarray: index of queried data
        """
        return scores.argsort()[::-1][:k]

    def cal_scores(self, probs: np.ndarray, steps: int = 20) -> np.ndarray:  # B,C,H,W
        """Calculate score by probability.

        Args:
            probs (np.ndarray): probability
            steps (int, optional): num of steps. Defaults to 20.

        Returns:
            np.ndarray: scores
        """
        num_classes = self.model_params["NUM_CLASSES"]
        outputs = torch.FloatTensor(
            probs.shape[0], num_classes, probs.shape[2], probs.shape[3]
        ).fill_(0)

        with torch.no_grad():
            for _ in range(steps):
                outputs[:, :, :, :] += probs

        outputs /= steps
        scores = []
        for i in range(probs.shape[0]):
            entropy_map = torch.FloatTensor(probs.shape[2], probs.shape[3]).fill_(0)
            for c in range(num_classes):
                entropy_map = entropy_map - (
                    outputs[i, c, :, :] * torch.log2(outputs[i, c, :, :] + 1e-12)
                )
            score = np.mean(-np.nansum(entropy_map.cpu().numpy(), axis=0))
            scores.append(score)

        return np.array(scores)



from typing import List, Tuple

import numpy as np
from torch.utils.data import Dataset

from .entropy_sampling import EntropySampling
from .least_confidence import LeastConfidence
from .margin_sampling import MarginSampling
from .strategy import Strategy


def get_uncertain_samples_idxs(probs: np.ndarray, n: int, criteria: str) -> np.ndarray:
    """Get the K most informative samples based on the criteria

    Args:
        probs (np.ndarray): prediction probability, (B,C,H,W)
        n (int): num of data to query
        criteria (str): calculation method,
                        `lc`: LeastConfidence,
                        `ms`: MarginSampling,
                        `en`: EntropySampling,

    Raises:
        ValueError: if criteria is not available

    Returns:
        np.ndarray: index of queried data
    """
    if criteria == "lc":
        scores = LeastConfidence.cal_scores(probs)
        topk_idxs = LeastConfidence.get_topk_idxs(scores, n)
    elif criteria == "ms":
        scores = MarginSampling.cal_scores(probs)
        topk_idxs = MarginSampling.get_topk_idxs(scores, n)
    elif criteria == "en":
        scores = EntropySampling.cal_scores(probs)
        topk_idxs = EntropySampling.get_topk_idxs(scores, n)
    else:
        raise ValueError("criteria {} not found !".format(criteria))
    return topk_idxs


def get_high_confidence_samples(
    probs: np.ndarray, delta: float, k: int
) -> Tuple[np.ndarray, np.ndarray]:
    """Select high confidence samples from unlabled data whose entropy is smaller than the threshold

    Args:
        probs (np.ndarray): prediction probability, (B,C,H,W)
        delta (float): threshold
        k (int): num of data to query

    Returns:
        Tuple[np.ndarray, np.ndarray]: index of queried data, mask
    """
    scores, entropy_maps = CealSampling.cal_scores(probs)
    topk_idxs = CealSampling.get_topk_idxs(scores, k)
    pred_class = np.argmax(probs, axis=1)
    for idx, entropy_map in entropy_maps.items():
        entropy_map = np.argmax(entropy_map, axis=0)
        qualified_area = entropy_map <= delta
        if np.any(qualified_area):
            pred_class[idx][qualified_area is False] = 255
    mask = pred_class[topk_idxs]
    return topk_idxs, mask


class CealSampling(Strategy):
    """CEAL Sampling Class.

    Args:
        pool_images (List[str]): List of pool image paths.
        pool_labels (List[str]): List of pool label paths.
        val_images (List[str]): List of validation image paths.
        val_labels (List[str]): List of validation label paths.
        test_images (List[str]): List of test image paths.
        test_labels (List[str]): List of test label paths.
        idxs_lb (np.ndarray): Array of bool type to record labeled data.
        model_params (dict): Model parameters.
                            e.g. model_params = {
                                    "MODEL_NAME": MODEL_NAME,
                                    "ENCODER": ENCODER,
                                    "ENCODER_WEIGHTS": ENCODER_WEIGHTS,
                                    "NUM_CLASSES": NUM_CLASSES,
                                }
        dataset (Dataset): Dataset class.
        dataset_params (dict): Dataset parameters.
                            e.g.     dataset_params = {
                                            "training_augmentation": get_training_augmentation(),
                                            "validation_augmentation": get_validation_augmentation(),
                                            "preprocessing": get_preprocessing(preprocessing_fn),
                                            "classes": CamvidDataset.CLASSES,
                                        }
    """

    def __init__(
        self,
        pool_images: List[str],
        pool_labels: List[str],
        val_images: List[str],
        val_labels: List[str],
        test_images: List[str],
        test_labels: List[str],
        idxs_lb: np.ndarray,
        model_params: dict,
        dataset: Dataset,
        dataset_params: dict,
        start_entropy_threshold: float = 0.0275,
        entropy_change_per_selection: float = 0.001815,
    ):
        super(CealSampling, self).__init__(
            pool_images,
            pool_labels,
            val_images,
            val_labels,
            test_images,
            test_labels,
            idxs_lb,
            model_params,
            dataset,
            dataset_params,
        )

        self.current_entropy_threshold = start_entropy_threshold
        self.entropy_change_per_selection = entropy_change_per_selection

    def query(self, n: int, criteria: str = "lc") -> List[int]:
        """Query data.

        Args:
            n (int): number of data to query.
            criteria (str): calculation method,
                            `lc`: LeastConfidence,
                            `ms`: MarginSampling,
                            `en`: EntropySampling,
        Returns:
            List[int]: Indices of queried data.
        """
        idxs_unlabeled = np.arange(self.n_pool)[
            ~self.idxs_lb
        ]  # reserve the index of unlabeled data
        probs = self.predict_prob(idxs_unlabeled)
        us_idxs = get_uncertain_samples_idxs(probs, n, criteria)
        us_idxs = idxs_unlabeled[us_idxs]  # idxs_queried: index in pool_images
        hcs_idx, _ = get_high_confidence_samples(
            probs, self.current_entropy_threshold, n
        )
        hcs_idx = idxs_unlabeled[hcs_idx]  # idxs_queried: index in pool_images
        idxs_queried = list(set(us_idxs) | set(hcs_idx))
        self.current_entropy_threshold -= self.entropy_change_per_selection
        return idxs_queried

    @staticmethod
    def get_topk_idxs(scores: np.ndarray, k: int) -> np.ndarray:
        """Get top k indices.

        Args:
            scores (np.ndarray): scores of batch data
            k (int): num of data to query

        Returns:
            np.ndarray: index of queried data
        """
        if isinstance(scores, list):
            scores = np.array(scores)
        return scores.argsort()[::-1][:k]

    @staticmethod
    def cal_scores(probs: np.ndarray) -> Tuple[np.ndarray, dict]:  # B,C,H,W
        """Calculate score by probability.

        Args:
            probs (np.ndarray): prediction probability

        Returns:
            Tuple[np.ndarray, dict]: scores of images, entropy map
        """
        scores = []
        entropy_maps = {}
        for i in range(len(probs)):  # one img prob
            entropy = np.multiply(probs[i], np.log2(probs[i] + 1e-12))
            score = np.mean(-np.nansum(entropy, axis=0))
            scores.append(score)
            entropy_maps[i] = entropy
        return np.array(scores), entropy_maps
